import { EntityRepository, getConnection, getManager, Repository } from "typeorm";
import { CreateProductServiceDto } from "./dto/create-product-service.dto";
import { UpdateProductServiceDto } from "./dto/update-product-service.dto";
import { ProductService } from "./entities/product-service.entity";

@EntityRepository(ProductService)
export class ProductServiceRepository extends Repository<ProductService> {
    async createEntity(dto: CreateProductServiceDto): Promise<any | Error> {
        try {
            const entity = this.create({
                retailerId: dto.retailerId,
                retailerName: dto.retailerName,
                productId: dto.productId,
                productName: dto.productName,
                productSlug: dto.productSlug,
                image: dto.image,
                brandName: dto.brandName,
                category: dto.category,
                variants: dto.variants
            });
            await this.save(entity);
            return entity;
        } catch (e) {
            return e;
        }
    }

    async findAll(): Promise<any | Error> {
        try {
            return await this.find({
                order: {
                    createdAt: -1
                },
                where: {
                    isDeleted: false
                }
            });
        } catch (e) {
            return e;
        }
    }

    async findByFilterQuery(query: any): Promise<any | Error> {
        try {
            return await this.find({
                where: { ...query, isDeleted: false }
            });
        } catch (e) {
            return e;
        }
    }

    async findOneEntity(slug: string, retailerId: string): Promise<any | Error> {
        try {
            return await this.findOne({
                where: {
                    retailerId: retailerId,
                    productSlug: slug,
                    isDeleted: false
                }
            });
        } catch (e) {
            return e;
        }
    }

    async deleteEntity(id: string): Promise<boolean | Error> {
        try {
            const updateResult = await getConnection()
              .createQueryBuilder()
              .update(ProductService)
              .set({ isDeleted: true })
              .where("id = :id", { id: id })
              .execute();
            return updateResult.affected > 0;
        } catch (e) {
            return e;
        }
    }

    async updateProductCategory(
      id: string,
      dto: UpdateProductServiceDto
    ): Promise<ProductService | Error> {
        try {
            await this.update(id, {
                retailerId: dto.retailerId,
                retailerName: dto.retailerName,
                productId: dto.productId,
                productName: dto.productName,
                productSlug: dto.productSlug,
                image: dto.image,
                brandName: dto.brandName,
                category: dto.category,
                variants: dto.variants
            });
            return await this.findOne({ where: { id: id } });
        } catch (e) {
            return e;
        }
    }

    async filterProductEntity(
      page: number,
      take: number,
      retailerId: string,
      minPrice?: number,
      maxPrice?: number,
      category?: string,
      subCategory?: string,
      brandName?: string,
      strainType?: string,
      weight?: string,
      menuType?: string,
      potencyThcMin?: number,
      potencyThcMax?: number,
      potencyCbdMin?: number,
      potencyCbdMax?: number,
      sort?: string
    ): Promise<any | Error> {
        try {
            let whereQuery: any = ``;

            if (category) {
                whereQuery = `ps.category = '${category}'`;
            }

            if (subCategory) {
                if (whereQuery.length > 3) {
                    whereQuery += `and ps.subCategory = '${subCategory}'`;
                } else {
                    whereQuery += ` ps.subCategory = '${subCategory}'`;
                }
            }

            if (brandName) {
                if (whereQuery.length > 3) {
                    whereQuery += `and ps.brandName = '${brandName}'`;
                } else {
                    whereQuery += ` ps.brandName = '${brandName}'`;
                }
            }

            if (strainType) {
                if (whereQuery.length > 3) {
                    whereQuery += `and ps.strainType = '${strainType}'`;
                } else {
                    whereQuery += ` ps.strainType = '${strainType}'`;
                }
            }

            if (weight) {
                if (whereQuery.length > 3) {
                    whereQuery += `and pv.weight = '${weight}'`;
                } else {
                    whereQuery += ` pv.weight = '${weight}'`;
                }
            }

            if (minPrice && maxPrice) {
                if (menuType === "RECREATIONAL") {
                    whereQuery += `and (pv.priceRec >= ${minPrice} and pv.priceRec <= ${maxPrice})`;
                } else {
                    whereQuery += `and (pv.priceMed >= ${minPrice} and p.priceMed <= ${maxPrice})`;
                }
            }

            if (potencyCbdMin && potencyCbdMax) {
                whereQuery += `and (ppc.potencyCbdRange >= ${potencyCbdMin} and ppc.potencyCbdRange <= ${potencyCbdMax})`;
            }

            if (potencyThcMin && potencyThcMax) {
                whereQuery += `and (ppt.potencyThcRange >= ${potencyThcMin} and ppt.potencyThcRange <= ${potencyThcMax})`;
            }


            if (whereQuery.length > 3) {
                whereQuery += ` and ps.isDeleted=0 and ps.retailerId= '${retailerId}' `;
            } else {
                whereQuery += ` ps.isDeleted=0 and ps.retailerId= '${retailerId}' `;
            }

            if (whereQuery.substring(0, 3) === "and") {
                whereQuery = whereQuery.slice(4, whereQuery.length);
            }

            let conditionQuery = ``;

            if (!whereQuery) {
                conditionQuery = ``;
            } else {
                conditionQuery = ` WHERE ${whereQuery} `;
            }

            let offset = 0;


            if (page > 1) {
                offset = ((page - 1) * take);
            }


            let sql = `SELECT *
            FROM product_service AS ps
            JOIN product_effects AS pe ON ps.productId = pe.productId
            JOIN product_menu_types AS pmt ON ps.productId = pmt.productId
            JOIN product_variants As pv ON ps.productId = pv.productId
            JOIN product_potency_cbd As ppc ON ps.productId = ppc.productId
            JOIN product_potency_thc As ppt ON ps.productId = ppt.productId
            ${conditionQuery} GROUP BY ps.id limit ${offset}, ${take};`;

            let sqlTotal = `SELECT *
            FROM product_service AS ps
            JOIN product_effects AS pe ON ps.productId = pe.productId
            JOIN product_menu_types AS pmt ON ps.productId = pmt.productId
            JOIN product_variants As pv ON ps.productId = pv.productId
            JOIN product_potency_cbd As ppc ON ps.productId = ppc.productId
            JOIN product_potency_thc As ppt ON ps.productId = ppt.productId
            ${conditionQuery} GROUP BY ps.id;`;

            console.log(sql);
            //console.log(sqlTotal);
            const entityManager = getManager();
            const data = await entityManager.query(sql);
            const totalData = await entityManager.query(sqlTotal);
            const count = totalData.length;

            const pageCount = Math.ceil(count / take);

            const responseData = {
                page: page,
                limit: take,
                pageCount: pageCount,
                totalDataFound: count,
                data
            };

            return responseData;

        } catch (e) {
            return e;
        }
    }

    async getProductEffect(retailerId: string): Promise<any | Error> {
        try {
            let sql = `SELECT DISTINCT effects FROM product_effects
                WHERE retailerId='${retailerId}' AND isDeleted=0 ORDER BY effects ASC;`;

            console.log(sql);
            const entityManager = getManager();
            const data = await entityManager.query(sql);
            return data;
        } catch (e) {
            return e;
        }

    }

    async getProductMenuType(retailerId: string): Promise<any | Error> {
        try {
            let sql = `SELECT DISTINCT menuType FROM product_menu_types
                WHERE retailerId='${retailerId}' AND isDeleted=0;`;

            console.log(sql);
            const entityManager = getManager();
            const data = await entityManager.query(sql);
            return data;
        } catch (e) {
            return e;
        }

    }

    async getProductCategory(retailerId: string): Promise<any | Error> {
        try {
            let sql = `SELECT DISTINCT category FROM product_service
                WHERE retailerId='${retailerId}' AND isDeleted=0 ORDER BY category ASC;`;

            console.log(sql);
            const entityManager = getManager();
            const data = await entityManager.query(sql);
            return data;
        } catch (e) {
            return e;
        }

    }

    async getProductSubCategory(retailerId: string): Promise<any | Error> {
        try {
            let sql = `SELECT DISTINCT subCategory FROM product_service
                WHERE retailerId='${retailerId}' AND isDeleted=0 ORDER BY subCategory ASC;`;

            console.log(sql);
            const entityManager = getManager();
            const data = await entityManager.query(sql);
            return data;
        } catch (e) {
            return e;
        }

    }

    async getProductBrandName(retailerId: string): Promise<any | Error> {
        try {
            let sql = `SELECT DISTINCT brandName FROM product_service
                WHERE retailerId='${retailerId}' AND isDeleted=0 ORDER BY brandName ASC;`;

            console.log(sql);
            const entityManager = getManager();
            const data = await entityManager.query(sql);
            return data;
        } catch (e) {
            return e;
        }

    }

    async getProductStrainType(): Promise<any | Error> {
        try {
            let sql = `SELECT DISTINCT strainType FROM product_service
                WHERE isDeleted=0 ORDER BY strainType ASC;`;

            const entityManager = getManager();
            const data = await entityManager.query(sql);
            return data;
        } catch (e) {
            return e;
        }

    }

    async getProductWeight(retailerId: string): Promise<any | Error> {
        try {
            let sql = `SELECT DISTINCT weight  FROM product_variants
                WHERE retailerId='${retailerId}' AND isDeleted=0 ORDER BY weight ASC;`;

            console.log(sql);
            const entityManager = getManager();
            const data = await entityManager.query(sql);
            return data;
        } catch (e) {
            return e;
        }

    }

    async deleteRowsFromDatabase(productIds: any): Promise<boolean> {
        const updateResult = await getConnection()
          .createQueryBuilder()
          .delete()
          .from(ProductService)
          .where("productId IN (:...ids)", { ids: productIds })
          .execute();

        return updateResult.affected > 0;
    }

    async createEntities(dtos: any): Promise<any | Error> {
        await getConnection()
          .createQueryBuilder()
          .insert()
          .into(ProductService)
          .values(dtos)
          .execute();
    }
}
