import { HttpException, HttpStatus, Injectable } from "@nestjs/common";
import { Constants } from "src/utils/constants";
import { CreateProductServiceDto } from "./dto/create-product-service.dto";
import { UpdateProductServiceDto } from "./dto/update-product-service.dto";
import { ProductServiceRepository } from "./product-service.repository";
import { ResponseUtils } from "src/utils/response.utils";
import { ProductService } from "./entities/product-service.entity";
import { getGraphQLClient } from "src/helper/graphql-client-helper";
import {
    fetchAllProductsForScheduler,
    fetchAllRetailers,
    fetchOneProduct,
    fetchProductsCount
} from "src/helper/dutchie-plus-query";
import { ElasticsearchService } from "@nestjs/elasticsearch";
import { SearchProductServiceDto } from "./dto/search-product-service.dto";
import { InjectRepository } from "@nestjs/typeorm";
import { ProductCategoryRepository } from "../product-category/product-category.repository";
import { urlSlug } from "src/helper/common-helper";
import { ProductEffectsRepository } from "../product-effects/product-effects.repository";
import { RetailerServiceRepository } from "../retailer-service/retailer-service.repository";
import { integer } from "aws-sdk/clients/cloudfront";

@Injectable()
export class ProductServiceService {
    constructor(
        private readonly repository: ProductServiceRepository,
        private readonly esService: ElasticsearchService,
        @InjectRepository(ProductCategoryRepository)
        private readonly productCatRepo: ProductCategoryRepository,
        @InjectRepository(ProductEffectsRepository)
        private readonly productEffectRepo: ProductEffectsRepository,
        @InjectRepository(RetailerServiceRepository)
        private readonly retailerServiceRepo: RetailerServiceRepository

    ) {
    }


    async processProductData(productData: any, retailerId: string, retailerName: string): Promise<any | Error> {
        try {
            const [
                productService,
                productCategory,
                productEffects
            ] = await Promise.all([
                this.repository.findByFilterQuery({}),
                this.productCatRepo.findAll(),
                this.productEffectRepo.findByFilterQuery({})
            ]);

            const productServiceDatasetIds = new Set(productService.map((item) => item.productId));
            const productMapService: any = new Map(productService.map(item => [item.productId, item]));
            const productDifferences = [];
            const differeneceDeleteIds = [];
            const categoriesArr = [];
            const effectsArr = [];

            await Promise.all(productData.menu.products.map(async (item) => {
                // product service
                if (productServiceDatasetIds.size > 0) {
                    if (!productServiceDatasetIds.has(item.id)) {
                        productDifferences.push(item);
                    } else {
                        // Now you can find a product by id in constant time:
                        const matchingItem = productMapService.get(item.id);
                        // const matchingItem = productService.find((item2) => item2.productId === item.id);
                        const brand = item.brand !== null ? item.brand.name : "";
                        if (
                            matchingItem &&
                            (matchingItem.brandName !== brand ||
                                matchingItem.image !== item.image ||
                                matchingItem.productName !== item.name ||
                                matchingItem.productSlug !== item.slug ||
                                matchingItem.category !== item.category ||
                                matchingItem.subCategory !== item.subcategory ||
                                matchingItem.strainType !== item.strainType)
                        ) {
                            differeneceDeleteIds.push(item.id);
                            productDifferences.push(item);
                        }
                    }
                } else {
                    productDifferences.push(item);
                }

                // all categories 
                categoriesArr.push(item.category);
                // all effects
                effectsArr.push(...item.effects);
            }));


            // if any update occers delete 
            if (differeneceDeleteIds.length > 0) {
                await this.repository.deleteRowsFromDatabase(differeneceDeleteIds);
            }

            // create entities 
            console.log(retailerName, ">", " New Product = ", productDifferences.length);
            const createProductServiceDtos = productDifferences.map((item) => {
                const brand = item.brand !== null ? item.brand.name : "";
                return {
                    retailerId,
                    retailerName,
                    productId: item.id,
                    productName: item.name,
                    productSlug: item.slug,
                    image: item.image,
                    brandName: brand,
                    category: item.category,
                    subCategory: item.subcategory,
                    strainType: item.strainType,
                    variants: item.variants.length
                };
            });

            if (createProductServiceDtos.length > 0) {
                await this.repository.createEntities(createProductServiceDtos);
            }

            // check effects
            const effects = Array.from(new Set(effectsArr));

            if (Object.keys(productEffects).length > 0) {
                const difference = effects.filter((effect) =>
                    !productEffects.some((e) => e.effects === effect)
                );
                console.log("Product Effects Difference = ", difference);
                // if difference found then create entities with difference product effects
                const createProductEffectsDtos = difference.map((item) => {
                    return {
                        effects: item
                    };
                });

                if (createProductEffectsDtos.length > 0) {
                    await this.productEffectRepo.createEntities(createProductEffectsDtos);
                }
            } else {
                // if product effects table is empty in our local database then create all the effect in local
                const createProductEffectsDtos = effects.map((item) => {
                    return {
                        effects: item
                    };
                });

                if (createProductEffectsDtos.length > 0) {
                    await this.productEffectRepo.createEntities(createProductEffectsDtos);
                }
            }

            // check category
            const categories = Array.from(new Set(categoriesArr));

            if (Object.keys(productCategory).length > 0) {
                const difference = categories.filter((category) =>
                    !productCategory.some((c) => c.categoryName === category)
                );
                console.log("Product Categories Difference = ", difference);
                // if difference found then create entities with difference product category
                const createProductCategoryDtos = difference.map((item) => {
                    return {
                        categoryName: item,
                        urlSlug: urlSlug(item)
                    };
                });

                if (createProductCategoryDtos.length > 0) {
                    await this.productCatRepo.createEntities(createProductCategoryDtos);
                }
            } else {
                // if product categories table is empty in our local database then create all the category in local
                const createProductCategoryDtos = categories.map((item) => {
                    return {
                        categoryName: item,
                        urlSlug: urlSlug(item)
                    };
                });

                if (createProductCategoryDtos.length > 0) {
                    await this.productCatRepo.createEntities(createProductCategoryDtos);
                }
            }

        } catch (e) {
            console.log(e);
            return e;
        }
    }


    async createUpdateDutchiePlusProductsAndCategory() {
        try {
            console.log(">>>Run Create or Update Dutchie Plus Products Scheduler<<<");

            // Get GraphQL client
            const graphQLClient = getGraphQLClient();

            // Fetch all retailers
            const query = fetchAllRetailers();
            const data = await graphQLClient.request(query);
            const retailers = data.retailers;

            // Check if retailers exist
            if (retailers.length > 0) {
                // Create an array of promises for fetching product data
                console.log("Start Fetching product count Parallel  =", new Date());
                const fetchProductsCountObject = fetchProductsCount();
                const retailerListPromises = retailers.map((retailer) =>
                    graphQLClient.request(fetchProductsCountObject, { retailerId: retailer.id })
                        .then(response => {
                            if (response && response.menu && response.menu.productsCount != null) {

                                return {
                                    retailerId: retailer.id,
                                    retailerName: retailer.name,
                                    pagination: {
                                        offset: 0,
                                        limit: response.menu.productsCount ? parseInt(response.menu.productsCount) : 0
                                    }
                                };
                            }
                        })
                );

                let allRetailerProductCount = await Promise.all(retailerListPromises);
                allRetailerProductCount = allRetailerProductCount.filter((data) => data.pagination.limit > 0);
                console.log("End Fetching product count Parallel =", new Date());


                console.log("Start Fetching product Parallel End =", new Date());
                const fetchProductsQuery = fetchAllProductsForScheduler();
                const fetchProductPromises = allRetailerProductCount.map(variable =>
                    graphQLClient.request(fetchProductsQuery, variable).then(response => {
                        if (response && response.menu && response.menu.products) {
                            return {
                                retailerId: variable.retailerId,
                                retailerName: variable.retailerName,
                                productData: response
                            };
                        }
                    })
                );
                console.log("End Fetching product Parallel End =", new Date());

                const allProductData = await Promise.all(fetchProductPromises);
                //console.log(allProductData)
                for (const data of allProductData) {
                    console.log("Send for Processing", new Date());
                    await this.processProductData(data.productData, data.retailerId, data.retailerName);
                    console.log("End processing =", new Date());
                }
                console.log(">>>End Create or Update Dutchie Plus Products Scheduler<<<");
            } else {
                console.log("Retailers Not Found!");
            }
        } catch (e) {
            console.log(e);
            return e;
        }
    }

    async productSynchronize(): Promise<any | Error> {
        try {
            await this.createUpdateDutchiePlusProductsAndCategory();
            return ResponseUtils.successResponseHandler(200, "Product Synchronize completed Successfully!");
        } catch (e) {
            console.log(e);
            return e;
        }
    }

    async create(dto: CreateProductServiceDto): Promise<any | Error> {
        const data = await this.repository.createEntity(dto);
        if (!data) {
            throw new HttpException(Constants.CREATE_FAILED, HttpStatus.BAD_REQUEST);
        }
        return ResponseUtils.successResponseHandler(
            201,
            "Products created successfully",
            "data",
            data
        );
    }

    async filterList(retailerId: string): Promise<any | Error> {
        try {
            const categoryData = await this.repository.getProductCategory(retailerId);
            const subCategoryData = await this.repository.getProductSubCategory(retailerId);
            const weightData = await this.repository.getProductWeight(retailerId);
            const effectsData = await this.repository.getProductEffect(retailerId);
            const menuTypeData = await this.repository.getProductMenuType(retailerId);
            const brandData = await this.repository.getProductBrandName(retailerId);
            const strainTypeData = await this.repository.getProductStrainType();

            const response = {
                categoryData: categoryData.map(item => item.category),
                subCategoryData: subCategoryData.map(item => item.subCategory),
                weightData: weightData.map(item => item.weight),
                effectsData: effectsData.map(item => item.effects),
                menuTypeData: menuTypeData.map(item => item.menuType),
                brandData: brandData.map(item => item.brandName),
                strainTypeData: strainTypeData.map(item => item.strainType)
            };

            return ResponseUtils.successResponseHandler(
                200,
                "Products filter list fetched successfully",
                "data",
                response
            );
        } catch (error) {
            console.log(error);
            return error;
        }
    }

    async filterProduct(
        page: number,
        limit: number,
        retailerId: string,
        minPrice?: number,
        maxPrice?: number,
        category?: string,
        subCategory?: string,
        brandName?: string,
        strainType?: string,
        weight?: string,
        menuType?: string,
        potencyThcMin?: number,
        potencyThcMax?: number,
        potencyCbdMin?: number,
        potencyCbdMax?: number,
        sort?: string
    ): Promise<any | Error> {
        const data = await this.repository.filterProductEntity(
            page,
            limit,
            retailerId,
            minPrice,
            maxPrice,
            category,
            subCategory,
            brandName,
            strainType,
            weight,
            menuType,
            potencyThcMin,
            potencyThcMax,
            potencyCbdMin,
            potencyCbdMax,
            sort
        );
        if (!data) {
            throw new HttpException(Constants.NOT_FOUND, HttpStatus.NOT_FOUND);
        }
        return ResponseUtils.successResponseHandler(
            200,
            "Products list fetched successfully",
            "data",
            data
        );
    }

    async findAll(): Promise<ProductService[] | Error> {
        try {
            const data = await this.repository.findAll();
            if (!data) {
                throw new HttpException(Constants.NOT_FOUND, HttpStatus.NOT_FOUND);
            }
            return ResponseUtils.successResponseHandler(
                200,
                "Products list fetched successfully",
                "data",
                data
            );
        } catch (error) {
            console.log(error);
            return error;
        }
    }

    async createUpdateDutchiePlusNewProducts(retailerId: string, retailerName: string) {
        try {
            console.log(">>>Run Create or Update Dutchie Plus New Products Scheduler using slug<<<");
            // Get GraphQL client
            const graphQLClient = getGraphQLClient();
            // Create an array of promises for fetching product data
            console.log("Start Fetching product count Parallel  =", new Date());
            const fetchProductsCountObject = fetchProductsCount();
            let allRetailerProductCount = await graphQLClient.request(fetchProductsCountObject, { retailerId: retailerId })
                .then(response => {
                    if (response && response.menu && response.menu.productsCount != null) {
                        return {
                            retailerId: retailerId,
                            retailerName: retailerName,
                            pagination: {
                                offset: 0,
                                limit: response.menu.productsCount ? parseInt(response.menu.productsCount) : 0
                            }
                        };
                    }
                })
            // console.log(allRetailerProductCount)
            console.log("End Fetching product count Parallel =", new Date());
            if (allRetailerProductCount.pagination.limit > 0) {
                console.log("Start Fetching product Parallel End =", new Date());
                const fetchProductsQuery = fetchAllProductsForScheduler();
                const fetchProductPromises = await graphQLClient.request(fetchProductsQuery, allRetailerProductCount).then(response => {
                    if (response && response.menu && response.menu.products) {
                        return {
                            retailerId: allRetailerProductCount.retailerId,
                            retailerName: allRetailerProductCount.retailerName,
                            productData: response
                        };
                    }
                })
                console.log("End Fetching product Parallel End =", new Date());

                console.log("Send for Processing", new Date());
                await this.processProductData(fetchProductPromises.productData, fetchProductPromises.retailerId, fetchProductPromises.retailerName);
                console.log("End processing =", new Date());
                console.log(">>>End Create or Update Dutchie Plus New Products Scheduler using slug<<<<<<");
            }
        } catch (e) {
            console.log(e);
            return e;
        }
    }

    async findOne(slug: string, retailerId: string, countHitAPI: integer = 0): Promise<any | Error> {
        try {
            const data = await this.repository.findOneEntity(slug, retailerId);
            if (data) {
                const variables = {
                    retailerId: data.retailerId,
                    productId: data.productId
                };
                const graphQLClient = getGraphQLClient();
                const query = fetchOneProduct();
                const findOnedata = await graphQLClient.request(query, variables);
                return findOnedata;
            } else {

                if (countHitAPI < 2) {
                    console.log('countHitAPI =', countHitAPI)
                    console.log(">>Fetch Products From Dutchie using URL START<<<<<<<<<<<<");

                    const retailerService = await this.retailerServiceRepo.findOne({ retailerId: retailerId })
                    if (retailerService) {
                        await this.createUpdateDutchiePlusNewProducts(retailerId, retailerService.retailerName);
                        countHitAPI += 1;
                        console.log(">>Fetch Products From Dutchie using URL END>>>>>>>>>>>>>>>");
                        return await this.findOne(slug, retailerId, countHitAPI);

                    }


                }

            }
        } catch (error) {
            console.log(error);
            return error;
        }
    }

    async findByURLSlug(slug: string): Promise<any | Error> {
        try {
            const data = await this.repository.findOne({
                where: {
                    productSlug: slug,
                    isDeleted: false
                }
            });
            if (!data) {
                throw new HttpException(Constants.NOT_FOUND, HttpStatus.NOT_FOUND);
            }
            return ResponseUtils.successResponseHandler(
                200,
                "Data fetched successfully",
                "data",
                data
            );
        } catch (error) {
            console.log(error);
            return error;
        }
    }

    async update(
        id: string,
        dto: UpdateProductServiceDto
    ): Promise<ProductService | Error> {
        const data = await this.repository.updateProductCategory(id, dto);
        if (!data) {
            throw new HttpException(Constants.UPDATE_FAILED, HttpStatus.BAD_REQUEST);
        }
        return ResponseUtils.successResponseHandler(
            200,
            "Products updated successfully",
            "data",
            data
        );
    }

    async remove(id: string): Promise<boolean | Error> {
        const data = await this.repository.deleteEntity(id);
        if (!data) {
            throw new HttpException(Constants.NOT_FOUND, HttpStatus.BAD_REQUEST);
        }
        return ResponseUtils.buildDeletedData(
            data,
            200,
            "Products deleted successfully"
        );
    }

    //#region Elasticsearch

    public async indexIsExists(index: string) {
        return await this.esService.indices.exists({ index });
    }

    public async deleteIndex(index: string) {
        const isExists = await this.indexIsExists(index);
        if (isExists) {
            await this.esService.indices.delete({ index });
        }
    }

    private async deleteDocumentById(index: string, id: string): Promise<any> {
        try {
            const response = await this.esService.delete({ index: index, id: id });
            console.log(`Document with _id '${id}' deleted successfully.`);
            return response;
        } catch (error) {
            console.error(`Error deleting document with _id '${id}':`, error);
            throw error;
        }
    }

    public async createIndex(index: string, body: any) {
        const isExists = await this.indexIsExists(index);
        if (!isExists) {
            const data = await this.esService.indices.create({ index, body });
            return data.acknowledged;
        }
        return false;
    }

    public async indexData(index: string, body: any) {
        return await this.esService.index({ index, body });
    }

    async fetchAllProductsForElasticSearch() {
        try {
            console.log(">>>Run Fetch All Products For Elastic Search<<<");

            // Get GraphQL client
            const graphQLClient = getGraphQLClient();

            // Fetch all retailers
            const query = fetchAllRetailers();
            const data = await graphQLClient.request(query);
            const retailers = data.retailers;

            // Check if retailers exist
            if (retailers.length > 0) {
                // Create an array of promises for fetching product data
                console.log("Start Fetching product count Parallel  =", new Date());
                const fetchProductsCountObject = fetchProductsCount();
                const retailerListPromises = retailers.map((retailer) =>
                    graphQLClient.request(fetchProductsCountObject, { retailerId: retailer.id })
                        .then(response => {
                            if (response && response.menu && response.menu.productsCount != null) {
                                return {
                                    retailerId: retailer.id,
                                    retailerName: retailer.name,
                                    pagination: {
                                        offset: 0,
                                        limit: response.menu.productsCount ? parseInt(response.menu.productsCount) : 0
                                    }
                                };
                            }
                        })
                );

                let allRetailerProductCount = await Promise.all(retailerListPromises);
                allRetailerProductCount = allRetailerProductCount.filter((data) => data.pagination.limit > 0);
                console.log("End Fetching product count Parallel =", new Date());

                console.log("Start Fetching product Parallel =", new Date());
                const fetchProductsQuery = fetchAllProductsForScheduler();
                const fetchProductPromises = allRetailerProductCount.map(variable =>
                    graphQLClient.request(fetchProductsQuery, variable).then(response => {
                        if (response && response.menu && response.menu.products) {
                            return {
                                retailerId: variable.retailerId,
                                retailerName: variable.retailerName,
                                productData: response
                            };
                        }
                    })
                );
                console.log("End Fetching product Parallel =", new Date());

                // console.table(allRetailerProductCount);

                const allProductData = await Promise.all(fetchProductPromises);
                const products = [];
                for (const data of allProductData) {
                    const product = data.productData.menu.products.map(x => {
                        const { id, brand, name, slug, variants, ...restData } = x;
                        const brandName = brand?.name ?? "";
                        const brandNameAndProductNameAndCategoryName = name.trim().startsWith(brandName.trim())
                            ? `${name} ${x.category}` : `${brandName} ${name} ${x.category}`;

                        return {
                            ...restData,
                            productId: id,
                            productName: name,
                            productSlug: slug,
                            brandName,
                            variants: variants?.length ?? 0,
                            retailerId: data.retailerId,
                            retailerName: data.retailerName,
                            brandNameAndProductNameAndCategoryName
                        };
                    });
                    products.push(...product);
                }
                console.log(">>>End Fetch All Products For Elastic Search<<<");
                // console.table(products.length);
                return products;
            } else {
                console.log("Retailers Not Found!");
            }
        } catch (e) {
            console.log(e);
            return e;
        }
    }

    public async indexBulkData(indexValue: string) {
        // const res: any = await this.findAll();
        // const products: Array<any> = res?.data;
        const products: any = await this.fetchAllProductsForElasticSearch();

        await this.deleteIndex(indexValue);

        const body = {
            settings: {
                analysis: {
                    filter: {
                        autocomplete_filter: {
                            type: "edge_ngram",
                            min_gram: 1,
                            max_gram: 15
                        }
                    },
                    analyzer: {
                        autocomplete: {
                            type: "custom",
                            tokenizer: "standard",
                            filter: ["lowercase", "autocomplete_filter"]
                        }
                    }
                }
            },

            mappings: {
                properties: {
                    category: {
                        type: "text",
                        fields: {
                            keyword: {
                                type: "keyword",
                                ignore_above: 256
                            }
                        }
                    },
                    subcategory: {
                        type: "text",
                        fields: {
                            keyword: {
                                type: "keyword",
                                ignore_above: 256
                            }
                        }
                    },
                    strainType: {
                        type: "text",
                        fields: {
                            keyword: {
                                type: "keyword",
                                ignore_above: 256
                            }
                        }
                    },
                    image: {
                        type: "text",
                        fields: {
                            keyword: {
                                type: "keyword",
                                ignore_above: 512
                            }
                        }
                    },
                    productId: {
                        type: "text",
                        fields: {
                            keyword: {
                                type: "keyword",
                                ignore_above: 256
                            }
                        }
                    },
                    productName: {
                        type: "text",
                        fields: {
                            keyword: {
                                type: "keyword",
                                ignore_above: 256
                            }
                        }
                    },
                    brandName: {
                        type: "text",
                        fields: {
                            keyword: {
                                type: "keyword",
                                ignore_above: 256
                            }
                        }
                    },
                    brandNameAndProductNameAndCategoryName: {
                        type: "text",
                        analyzer: "autocomplete",
                        search_analyzer: "standard",
                        fields: {
                            keyword: {
                                type: "keyword",
                                ignore_above: 512
                            }
                        }
                    },
                    productSlug: {
                        type: "text",
                        fields: {
                            keyword: {
                                type: "keyword",
                                ignore_above: 256
                            }
                        }
                    },
                    retailerId: {
                        type: "text",
                        fields: {
                            keyword: {
                                type: "keyword",
                                ignore_above: 256
                            }
                        }
                    },
                    retailerName: {
                        type: "text",
                        fields: {
                            keyword: {
                                type: "keyword",
                                ignore_above: 256
                            }
                        }
                    },
                    variants: {
                        type: "long"
                    }
                }
            }
        };
        await this.createIndex(indexValue, body);
        console.log(">>>Start Flat Map<<<");
        const operations = products.flatMap((doc) => [
            { index: { _index: indexValue } },
            doc
        ]);
        console.log(">>>End Flat Map<<<");

        console.log(">>>Start Bulk Insert<<<");
        await this.esService.bulk({ refresh: true, operations });
        console.log(">>>End Bulk Insert<<<");

        return ResponseUtils.successResponseHandler(200, "Elastic DataBase Update Successfully.");
    }

    public async search(index: string, searchProductServiceDto: SearchProductServiceDto) {
        const { search, customField, page, limit, retailerId } = searchProductServiceDto;

        let exactsMatch = [];
        if (retailerId && retailerId != "") {
            const retailersMatch = {
                match_phrase: {
                    retailerId: retailerId.toLowerCase()
                }
            };
            exactsMatch = [...exactsMatch, retailersMatch];
        }

        let fields: Array<string> = ["brandNameAndProductNameAndCategoryName"];
        if (customField && customField?.length) {
            fields = customField;
        }

        const result = await this.esService.search<any>({
            index,
            body: {
                query: {
                    bool: {
                        must: [
                            ...exactsMatch,
                            {
                                match: {
                                    brandNameAndProductNameAndCategoryName: {
                                        query: search,
                                        fuzziness: "AUTO",
                                        fuzzy_transpositions: false,
                                        prefix_length: 3
                                    }
                                }
                            }
                        ]
                    }
                },
                highlight: {
                    fields: {
                        brandNameAndProductNameAndCategoryName: {
                            pre_tags: ["<b>"],
                            post_tags: ["</b>"]
                        }
                    }
                }
            },
            // from: page && page > 0 ? page - 1 : 0,
            size: limit && limit > 0 ? limit : 1000
        });
        const hits = result.hits.hits;
        const pageCount: any = result.hits.total;

        const products = hits.map((item: any) => {
            // console.log(item._source.retailerId == retailerId);
            return {
                ...item._source,
                highlight: item?.highlight?.brandNameAndProductNameAndCategoryName?.[0]
            };
        });
        const data = { products, total: (pageCount?.value ?? pageCount) };

        if (!data) {
            throw new HttpException(Constants.NOT_FOUND, HttpStatus.NOT_FOUND);
        }
        return ResponseUtils.successResponseHandler(
            200,
            "Products list fetched successfully",
            "data",
            data
        );
    }

    public async searchDetails(index: string, searchProductServiceDto: SearchProductServiceDto) {
        const { search, customField, page, limit, retailerId } = searchProductServiceDto;

        let exactsMatch = [];
        if (retailerId && retailerId != "") {
            const retailersMatch = {
                match_phrase: {
                    retailerId: retailerId.toLowerCase()
                }
            };
            exactsMatch = [...exactsMatch, retailersMatch];
        }

        let fields: Array<string> = ["brandNameAndProductNameAndCategoryName"];
        if (customField && customField?.length) {
            fields = customField;
        }

        const result = await this.esService.search<any>({
            index,
            body: {
                query: {
                    bool: {
                        must: [
                            ...exactsMatch,
                            {
                                match: {
                                    brandNameAndProductNameAndCategoryName: {
                                        query: search,
                                        fuzziness: "AUTO",
                                        fuzzy_transpositions: false,
                                        prefix_length: 3
                                    }
                                }
                            }
                        ]
                    }
                }
            },
            // from: page ? page : 0,
            size: 1000
        });

        const data = result.hits.hits;
        const start = page ? page : 0;
        const end = limit > 0 ? limit : 20;
        const hits = data.slice(start * end, (start + 1) * end);
        const pageCount: any = result.hits.total;
        let menu = {
            weights: [],
            brands: [],
            productsCount: pageCount?.value ?? pageCount,
            products: [],
            subCategory: []
        };
        if (hits.length > 0) {
            const graphQLClient = getGraphQLClient();
            let BrandHashSet = new Set();
            let subCategoryHashSet = new Set();
            let weightsHashSet = new Set();
            const allApi = [];

            for (let i = 0; i < hits.length; i++) {
                const index_id = hits[i]._id;
                const productId = hits[i]._source.productId;
                const variables = { retailerId, productId };
                const query = fetchOneProduct();
                const value = graphQLClient.request(query, variables);
                allApi.push(value.catch((error: any) => {
                    // console.error(error);
                    // this.deleteDocumentById(index, index_id);
                    return Promise.resolve();
                }));
            }
            // const results = await Promise.all(allApi);
            await Promise.all(allApi)
                .then((results) => {
                    for (let j = 0; j < results.length; j++) {
                        if (results?.[j]?.product) {
                            menu.products.push(results[j].product);
                            results[j].product.variants.forEach((x: { option: any; }) => {
                                if (!weightsHashSet.has(x.option)) {
                                    weightsHashSet.add(x.option);
                                    menu.weights.push(x.option);
                                }
                            });
                            if (!BrandHashSet.has(results[j].product.brand.id)) {
                                BrandHashSet.add(results[j].product.brand.id);
                                menu.brands.push(results[j].product.brand);
                            }
                            if (!subCategoryHashSet.has(results[j].product.subcategory)) {
                                subCategoryHashSet.add(results[j].product.subcategory);
                                menu.subCategory.push(results[j].product.subcategory);
                            }
                        }
                    }
                })
                .catch(error => {
                    console.log(error);
                })
                .finally(() => {
                    return { menu };
                });
        }

        return { menu };
    }

    //#endregion Elasticsearch
}
